DungeonMate = LibStub("AceAddon-3.0"):NewAddon("DungeonMate","AceConsole-3.0","AceEvent-3.0")
--[[ UPVALUES ]]--
local lgt = LibStub("LibGroupTalents-1.0")
local dm_panel, plist, noteframe -- Frames
local ignorelist = {} -- Periodically updated list of every ignore on the character. UPDATE BEFORE USING.
local groupcache = {} -- Data structure of everyone in the group
local UpdatePartyDisplay -- PROTOYPE so that closures can grab the value.
local roleicons = "Interface\\LFGFrame\\UI-LFG-ICON-PORTRAITROLES.blp"
local dmicon = "Interface\\Icons\\Spell_Nature_StormReach"

----------[[ CATEGORY: MISCELLANEOUS ]]----------

local function UpdateIgnorelist()
	ignorelist = {} -- flush out the old
	local i = GetNumIgnores()
	while i > 0 do
		ignorelist[GetIgnoreName(i)]=true
		i = i - 1
	end
end

-- This function NOT USED at this time. Will probably be added later as command line option.
local function FullGroupReset()
	groupcache = {}
	UpdatePartyDisplay()
end

----------[[ CATEGORY: PERSISTANCE ]]----------

local function preparePDBRecord(name)
	if not DungeonMate.db.realm.pdb[name] then
		DungeonMate.db.realm.pdb[name] = {
			["favorite"] = false,
			["note"] = nil,
			["lastseentime"] = nil,
			["lastseenzone"] = nil,
		}
	end
	return DungeonMate.db.realm.pdb[name]
end

local function GetLastSeen(name)
	local record = preparePDBRecord(name)
	return record.lastseentime, record.lastseenzone
end

local function GetNote(name)
	return preparePDBRecord(name).note
end

local function GetFavorite(name)
	return preparePDBRecord(name).favorite
end

----------[[ CATEGORY: DATA GENERATORS ]]----------

local function ClearPartyGroupStatus()
	local player
	for _,player in pairs(groupcache) do
		player.isInGroup = false
	end
end

local function UpdateGroupMember(unit,name)
	-- This function is intended to add information to a groupcache member that is an active party member
	-- In other words, it supplements other data gathering methods with Unit-based functions.
	local entry = groupcache[name] -- Gaurunteed by RegenerateGroupCache() to not be nil
	entry.name = name -- Needed for... well, frankly, for making one less upvalue.
	entry.isInGroup = true
	
	-- Roles --
	local isTank, isHeal, isDPS = UnitGroupRolesAssigned(unit)
	if isTank then
		entry.role = "tank"
	elseif isHeal then
		entry.role = "heal"
	elseif isDPS then
		entry.role = "dps"
	else
		entry.role = "no role"
	end
	
	-- Class --
	entry.classname, entry.class = UnitClass(unit) -- classname is localized, class is standardized
	
	-- Level --
	entry.level = UnitLevel(unit)
	
	-- Talents --
	local dominant, t1, t2, t3 = lgt:GetUnitTalentSpec(unit)
	if dominant then
		entry.spec = {
			["dominant"] = dominant,
			["t1"] = t1,
			["t2"] = t2,
			["t3"] = t3,
		}
	end
				
	-- GearGauge --
	if GearGauge then
		local rating, equip_level = GearGauge:Scan(unit)
		entry.GearGauge = {
			["rating"] = rating or "??",
			["equip_level"] = equip_level or "??"
		}
	end
	
	-- Last Seen --
	-- NOTE: We add these entries on "currentseentime" because we don't want them to appear in the "last seen" when they're
	-- only just being seen there now! Instead, once they've left the group, there will be a call to UpdateNonPartyMembers()
	-- Which will move the currentseentime over to the lastseentime. Nifty, huh?
	-- PROBLEM: What if that update call doesn't get made before the table is wiped? Hmmmmm. let's see if it comes up ever.
	entry.currentseentime = time()
	entry.currentseenzone = GetZoneText() -- TODO: Make this Instance Aware, and not update when outside of an instance!
end

local function UpdatePartyMembers()
	local n = GetNumPartyMembers()
	while n > 0 do
		local unit = "party" .. n
		local name, realm = UnitName(unit)
		if realm and realm ~= "" then name = name .. "-" .. realm end
		if name and name ~= UNKNOWNOBJECT then
			groupcache[name] = groupcache[name] or {}
			UpdateGroupMember(unit,name)
		end
		n = n - 1
	end
end

local function UpdateNonPartyMembers()
	-- NOTE: This function should only do things that CANT be done on current Party Members.
	local player
	for _,player in pairs(groupcache) do
		if not player.isInGroup then
		
			-- ROLE UPDATE: No role for non members!
			player.role = "no role"
			
			-- LAST SEEN: Take the "current seen" values and make them last seen!
			-- NOTE: This is actually data *generating*, not data *gathering*, which is not normal for this function.
			-- Fret not, the actual line info will be generated in UpdatePlayerCache
			local record = preparePDBRecord(player.name)
			record.lastseentime = player.currentseentime
			record.lastseenzone = player.currentseenzone
		end
	end
end

local function UpdatePlayerCache(player)
	-- This is the function that updates EVERY ONE of the members of groupcache, whether they are in a group or not.
	
	-- Favorite --
	player.isFavorite = GetFavorite(player.name)

	-- Ignored --
	if ignorelist[player.name] then -- I use this construct to ensure only 'true' or 'false' gets returned
		player.isIgnored = true
	else
		player.isIgnored = false
	end

	-- Note --
	player.note = GetNote(player.name) -- Could be nil or a note, never "" (I hope!)
	
	-- Last Seen--
	player.lastseentime, player.lastseenzone = GetLastSeen(player.name)
end

local function UpdateGroupCache()
	local player
	for _,player in pairs(groupcache) do
		UpdatePlayerCache(player)
	end
end

local function RemoveGroupRecord(name)
	-- Note: This function should NOT update the UI.
	groupcache[name] = nil
end
----------[[ CATEGORY: INFOLINES ]]----------

local infolineRecycleQueue = {}

local function recycleInfoline(line)
	line:Hide()
	tinsert(infolineRecycleQueue,line)
end

local function clearAllLines(playerframe)
	while #(playerframe.linebox.lines) > 0 do
		local temp = tremove(playerframe.linebox.lines)
		recycleInfoline(temp)
		playerframe.linebox:SetHeight(playerframe.linebox:GetHeight()-temp:GetHeight())
		playerframe:SetHeight(playerframe:GetHeight()-temp:GetHeight())
	end
end

local function updateInfoline(frame,text)
	frame.label:SetText(text)
end

local function manufactureInfoline(text,parent)
	local resultframe
	if #infolineRecycleQueue > 0 then
		resultframe = tremove(infolineRecycleQueue)
	else
		resultframe = CreateFrame("Frame",nil,parent)
		resultframe.label = resultframe:CreateFontString(nil,"OVERLAY","GameFontNormal")
		resultframe.label:SetFont("Fonts\\FRIZQT__.TTF", 10)
		resultframe.label:SetPoint("LEFT",resultframe,"LEFT",0,0)
		resultframe:SetHeight(15)

	end
	resultframe:SetParent(parent)
	updateInfoline(resultframe,text)
	resultframe:Show()
	return resultframe
end

local function insertInfoline(text, playerframe)
	local infoline = manufactureInfoline(text, playerframe.linebox)
	if #(playerframe.linebox.lines) > 0 then
		local parentf = playerframe.linebox.lines[#(playerframe.linebox.lines)]
		infoline:SetPoint("TOPLEFT",parentf,"BOTTOMLEFT")
		infoline:SetPoint("TOPRIGHT",parentf,"BOTTOMRIGHT")
	else
		infoline:SetPoint("TOPLEFT",playerframe.linebox,"TOPLEFT")
		infoline:SetPoint("TOPRIGHT",playerframe.linebox,"TOPRIGHT")
	end
	playerframe.linebox:SetHeight(playerframe.linebox:GetHeight()+infoline:GetHeight())
	playerframe:SetHeight(playerframe:GetHeight()+infoline:GetHeight())
	tinsert(playerframe.linebox.lines,infoline)
end

----------[[ CATEGORY: NOTE FRAME ]]----------

local function SetNote(name,note)
	local record = preparePDBRecord(name)
	if note and note == "" then note = nil end -- Blank notes mean "clear the note please"
	record.note = note
	UpdatePartyDisplay()
end

local function StartEditingNote(name)
	noteframe.nameplate:SetText(name)
	noteframe.editbox:SetText(preparePDBRecord(name).note or "")
	noteframe.editbox:SetScript("OnEnterPressed",function(frame)
		SetNote(name,noteframe.editbox:GetText())
		noteframe:Hide()
	end)
	noteframe.submit:SetScript("OnClick",function(frame)
		SetNote(name,noteframe.editbox:GetText())
		noteframe:Hide()
	end)
	noteframe:Show()
	noteframe.editbox:SetFocus()
end

local function initializeNoteEditFrame()
	noteframe = CreateFrame("Frame", nil, UIParent)
	noteframe:SetBackdrop({
		bgFile ="Interface/DialogFrame/UI-DialogBox-Background", 
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
		tile = true, tileSize = 16, edgeSize = 16, 
		insets = { left = 4, right = 4, top = 4, bottom = 4 }
	})
	noteframe:SetPoint("CENTER",UIParent,"CENTER")
	noteframe:SetHeight(100)
	noteframe:SetWidth(300)
	
	local title = noteframe:CreateFontString(nil,"OVERLAY","GameFontHighlightLarge")
	title:SetFont("Fonts\\FRIZQT__.TTF", 16)
	title:SetPoint("TOP",noteframe,"TOP",0,-10)
	title:SetText("Edit Note")
	title:Show()
	
	local nameplate = noteframe:CreateFontString(nil,"OVERLAY","GameFontNormal")
	nameplate:SetFont("Fonts\\FRIZQT__.TTF", 12)
	nameplate:SetPoint("TOP",title,"BOTTOM",0,0)
	nameplate:Show()
	
	local editbox = CreateFrame("EditBox",nil,noteframe,"InputBoxTemplate")
	editbox:SetPoint("CENTER",noteframe,"CENTER",0,0)
	editbox:SetHeight(20)
	editbox:SetWidth(200)
	editbox:Show()
	
	local submit = CreateFrame("Button",nil,noteframe,"UIPanelButtonTemplate")
	submit:SetPoint("CENTER",noteframe,"CENTER",0,-30)
	submit:SetText("Submit")
	submit:SetHeight(20)
	submit:SetWidth(100)
	submit:Show()
	
	local close = CreateFrame("Button",nil,noteframe,"UIPanelCloseButton")
	close:SetPoint("TOPRIGHT",noteframe,"TOPRIGHT")
	
	noteframe:Hide()
	noteframe.title = title
	noteframe.editbox = editbox
	noteframe.submit = submit
	noteframe.nameplate = nameplate
end

----------[[ CATEGORY: PLAYER FRAMES ]]----------

local playerRecycleQueue = {}

local function ToggleFavorite(name)
	local record = preparePDBRecord(name)
	record.favorite = not record.favorite
	UpdatePartyDisplay()
end

local function SetPlayerIgnoreFavoriteIndicators(frame,isIgnored,isFavorite)
	frame:SetBackdropColor(0,0,0,0.8)
	frame.linebox.buttonbar.ignorebutton:SetText("Ignore")
	frame.linebox.buttonbar.favoritebutton:SetText("Favorite")
	
	if isFavorite then
		frame:SetBackdropColor(0,0,1,0.3)
		frame.linebox.buttonbar.favoritebutton:SetText("Un-Favorite")
	end
	
	if isIgnored then
		frame:SetBackdropColor(1,0,0,0.3)
		frame.linebox.buttonbar.ignorebutton:SetText("Un-Ignore")
	end
end

local function recyclePlayer(player)
	player:Hide()
	clearAllLines(player)
	tinsert(playerRecycleQueue,player)
end

local function manufacturePlayerFrame(player)
	local resultframe -- Input player is a player info table. Output is a frame based on that.
	if #playerRecycleQueue > 0 then
		resultframe = tremove(playerRecycleQueue) -- ok, so it's a stack... big deal
	else
		resultframe = CreateFrame("Frame",nil,plist)
		resultframe:SetHeight(45) -- filled by the titlebar + buttonbar
		resultframe:SetBackdrop({ bgFile ="Interface\\Addons\\DungeonMate\\blank", 
		                          edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
		                          tile = true, tileSize = 16, edgeSize = 16, 
		                          insets = { left = 4, right = 4, top = 4, bottom = 4 }})

		
		resultframe.titlebar = CreateFrame("Button",nil,resultframe)
		resultframe.titlebar:SetPoint("TOPLEFT",resultframe,"TOPLEFT")
		resultframe.titlebar:SetPoint("TOPRIGHT",resultframe,"TOPRIGHT")
		resultframe.titlebar:SetHeight(20)
		resultframe.titlebar:Show()
		
		
		resultframe.titlebar.label = resultframe.titlebar:CreateFontString(nil,"OVERLAY","GameFontNormal")
		resultframe.titlebar.label:SetFont("Fonts\\FRIZQT__.TTF", 14)
		resultframe.titlebar.label:SetPoint("TOPLEFT",resultframe,"TOPLEFT",25,-5)
		
		resultframe.titlebar:EnableMouse(true)

		resultframe.titlebar.role = resultframe.titlebar:CreateTexture(nil,"OVERLAY")
		resultframe.titlebar.role:SetPoint("TOPLEFT",resultframe.titlebar,"TOPLEFT",5,-4)
		resultframe.titlebar.role:SetHeight(16)
		resultframe.titlebar.role:SetWidth(16)
		
		resultframe.titlebar.remove = CreateFrame("Button",nil,resultframe.titlebar,"UIPanelCloseButton")
		resultframe.titlebar.remove:SetHeight(30)
		resultframe.titlebar.remove:SetWidth(30)
		resultframe.titlebar.remove:SetPoint("TOPRIGHT",resultframe.titlebar,"TOPRIGHT",3,3)
		
		resultframe.linebox = CreateFrame("Frame",nil,resultframe)
		resultframe.linebox:SetPoint("TOPLEFT",resultframe.titlebar,"BOTTOMLEFT",25,0)
		resultframe.linebox:SetPoint("TOPRIGHT",resultframe.titlebar,"BOTTOMRIGHT",0,0)
		resultframe.linebox:SetHeight(20) -- filled up by the buttonbar below
		
		resultframe.linebox:Show()
		
		resultframe.linebox.buttonbar = CreateFrame("Frame",nil,resultframe.linebox)
		resultframe.linebox.buttonbar:SetPoint("BOTTOMLEFT",resultframe.linebox,"BOTTOMLEFT")
		resultframe.linebox.buttonbar:SetPoint("BOTTOMRIGHT",resultframe.linebox,"BOTTOMRIGHT")
		resultframe.linebox.buttonbar:SetHeight(20) -- fills up the default space of the linebox
		resultframe.linebox.buttonbar:Show()
		
		resultframe.linebox.buttonbar.notebutton = CreateFrame("Button",nil,resultframe.linebox.buttonbar,"UIPanelButtonTemplate")
		resultframe.linebox.buttonbar.notebutton:SetPoint("TOPLEFT",resultframe.linebox.buttonbar,"TOPLEFT")
		resultframe.linebox.buttonbar.notebutton:SetPoint("BOTTOMLEFT",resultframe.linebox.buttonbar,"BOTTOMLEFT")
		resultframe.linebox.buttonbar.notebutton:SetWidth(70)
		resultframe.linebox.buttonbar.notebutton:GetFontString():SetFont("Fonts\\FRIZQT__.TTF", 10)
		resultframe.linebox.buttonbar.notebutton:GetFontString():SetPoint("CENTER",resultframe.linebox.buttonbar.notebutton,"CENTER",0,1)
		resultframe.linebox.buttonbar.notebutton:Show();
		
		resultframe.linebox.buttonbar.ignorebutton = CreateFrame("Button",nil,resultframe.linebox.buttonbar,"UIPanelButtonTemplate")
		resultframe.linebox.buttonbar.ignorebutton:SetPoint("TOPLEFT",resultframe.linebox.buttonbar.notebutton,"TOPRIGHT")
		resultframe.linebox.buttonbar.ignorebutton:SetPoint("BOTTOMLEFT",resultframe.linebox.buttonbar.notebutton,"BOTTOMRIGHT")
		resultframe.linebox.buttonbar.ignorebutton:SetWidth(70)
		resultframe.linebox.buttonbar.ignorebutton:GetFontString():SetFont("Fonts\\FRIZQT__.TTF", 10)
		resultframe.linebox.buttonbar.ignorebutton:GetFontString():SetPoint("CENTER",resultframe.linebox.buttonbar.ignorebutton,"CENTER",0,1)
		resultframe.linebox.buttonbar.ignorebutton:Show();
		
		resultframe.linebox.buttonbar.favoritebutton = CreateFrame("Button",nil,resultframe.linebox.buttonbar,"UIPanelButtonTemplate")
		resultframe.linebox.buttonbar.favoritebutton:SetPoint("TOPLEFT",resultframe.linebox.buttonbar.ignorebutton,"TOPRIGHT")
		resultframe.linebox.buttonbar.favoritebutton:SetPoint("BOTTOMLEFT",resultframe.linebox.buttonbar.ignorebutton,"BOTTOMRIGHT")
		resultframe.linebox.buttonbar.favoritebutton:SetWidth(70)
		resultframe.linebox.buttonbar.favoritebutton:GetFontString():SetFont("Fonts\\FRIZQT__.TTF", 10)
		resultframe.linebox.buttonbar.favoritebutton:GetFontString():SetPoint("CENTER",resultframe.linebox.buttonbar.favoritebutton,"CENTER",0,1)
		resultframe.linebox.buttonbar.favoritebutton:Show();
		
		
		resultframe.linebox.lines = {}
		
		resultframe.titlebar:SetScript("OnClick",function(frame)
			if resultframe.linebox:IsShown() then
				resultframe:SetHeight(resultframe:GetHeight()-resultframe.linebox:GetHeight())
				plist:SetHeight(plist:GetHeight()-resultframe.linebox:GetHeight())
				resultframe.linebox:Hide()
			else
				resultframe:SetHeight(resultframe:GetHeight()+resultframe.linebox:GetHeight())
				plist:SetHeight(plist:GetHeight()+resultframe.linebox:GetHeight())
				resultframe.linebox:Show()
			end
		end)
	end
	
	resultframe:Show()
	return resultframe
end


local function updatePlayer(player, frame)
	-- This is the function that actually puts the data into the frame. ONLY this function should do
	-- that.
	frame.info = player -- export the original info, in case it helps anyone.
	
	if player.isInGroup then         -- Hide the remove button while the player is in the group
		frame.titlebar.remove:Hide() -- This button wouldn't do anything anyway
	else                             -- since the player would be regenerated instantly in the next
		frame.titlebar.remove:Show() -- update call, which would be triggered by the button press.
	end
	
	SetPlayerIgnoreFavoriteIndicators(frame,player.isIgnored,player.isFavorite)
	
	-- NOTE BUTTON HANDLING --
	if player.note then
		frame.linebox.buttonbar.notebutton:SetText("Edit Note")
	else
		frame.linebox.buttonbar.notebutton:SetText("Add Note")
	end
	
	if player.role == "tank" then
		frame.titlebar.role:SetTexture(roleicons)
		frame.titlebar.role:SetTexCoord(0, 19/64, 22/64, 41/64)
		frame.titlebar.role:Show()
	elseif player.role == "heal" then
		frame.titlebar.role:SetTexture(roleicons)
		frame.titlebar.role:SetTexCoord(20/64, 39/64, 1/64, 20/64)
		frame.titlebar.role:Show()
	elseif player.role == "dps" then
		frame.titlebar.role:SetTexture(roleicons)
		frame.titlebar.role:SetTexCoord(20/64, 39/64, 22/64, 41/64)
		frame.titlebar.role:Show()
	else
		frame.titlebar.role:SetTexture(dmicon)
		frame.titlebar.role:SetTexCoord(0,1,0,1)
		frame.titlebar.role:Show()
	end
	
	clearAllLines(frame)
	
	local classinfo = player.level .. " " .. player.classname
	
	if player.spec then
		classinfo = classinfo.." "..player.spec.t1.."/"..player.spec.t2.."/"..player.spec.t3
		classinfo = classinfo .. " " .. player.spec.dominant 
	end
	
	insertInfoline(classinfo, frame)
	
	-- GearGauge --
	if player.GearGauge then
		insertInfoline("GearGauge Rating: "..player.GearGauge.rating .." iLevel: " .. player.GearGauge.equip_level,frame)
	end
	
	-- Note --
	if player.note then
		insertInfoline("Note: ".. player.note,frame)
	end
	
	-- Last Seen --
	if player.lastseentime  and player.lastseenzone then
		insertInfoline("Last seen: " .. date("%m/%d/%y",player.lastseentime) .. " - " .. player.lastseenzone,frame)
	end
end

local function insertPlayer(player)
	-- Keep in mind, at this point "player" is an info table, not a frame
	if not plist then return end
	local frame = manufacturePlayerFrame(player)
	if #(plist.group) > 0 then
		-- Attach to the last player
		local last = plist.group[#(plist.group)]
		frame:SetPoint("TOPLEFT",last,"BOTTOMLEFT")
		frame:SetPoint("TOPRIGHT",last,"BOTTOMRIGHT")
	else
		-- Attach to the top of the player list
		frame:SetPoint("TOPLEFT",plist,"TOPLEFT")
		frame:SetPoint("TOPRIGHT",plist,"TOPRIGHT")
	end
	
	-- Now we have a frame and it's attached to the right spot. Next, we dress it up, give it the right info.
	
	frame.titlebar.label:SetText(player.name)
	frame.titlebar.remove:SetScript("OnCLick",function(frame)
		RemoveGroupRecord(player.name)
		UpdatePartyDisplay()
	end)
	frame.linebox.buttonbar.notebutton:SetScript("OnClick",function(frame) StartEditingNote(player.name) end)
	frame.linebox.buttonbar.ignorebutton:SetScript("OnClick",function(frame) AddOrDelIgnore(player.name) end)
	frame.linebox.buttonbar.favoritebutton:SetScript("OnClick",function(frame) ToggleFavorite(player.name) end)

	if player.class == "DEATHKNIGHT" then
		frame.titlebar.label:SetTextColor(0.77,0.12,0.23)
	elseif player.class == "DRUID" then
		frame.titlebar.label:SetTextColor(1.00,0.49,0.04)
	elseif player.class == "HUNTER" then
		frame.titlebar.label:SetTextColor(0.67,0.83,0.45)
	elseif player.class == "MAGE" then
		frame.titlebar.label:SetTextColor(0.41,0.80,0.94)
	elseif player.class == "PALADIN" then
		frame.titlebar.label:SetTextColor(0.96,0.55,0.73)
	elseif player.class == "PRIEST" then
		frame.titlebar.label:SetTextColor(1.00,1.00,1.00)
	elseif player.class == "ROGUE" then
		frame.titlebar.label:SetTextColor(1.00,0.96,0.41)
	elseif player.class == "SHAMAN" then
		frame.titlebar.label:SetTextColor(0.14,0.35,1.00)
	elseif player.class == "WARLOCK" then
		frame.titlebar.label:SetTextColor(0.58,0.51,0.79)
	elseif player.class == "WARRIOR" then
		frame.titlebar.label:SetTextColor(0.78,0.61,0.43)
	else 
		frame.titlebar.label:SetTextColor(0.5,0.5,0.5)
	end
	
	updatePlayer(player,frame)
	
	plist:SetHeight(plist:GetHeight()+frame:GetHeight())
	tinsert(plist.group,frame)
end

----------[[ CATEGORY: LIST FRAME ]]----------

local function clearParty()
	if not plist then return end
	while #(plist.group) > 0 do
		local temp = tremove(plist.group)
		temp:Hide()
		plist:SetHeight(plist:GetHeight()-temp:GetHeight())
		recyclePlayer(temp)
	end
end

local function BuildGroupDisplay()
	local player
	for _,player in pairs(groupcache) do
		insertPlayer(player)
	end
end

-- PROTOTYPED as local
function UpdatePartyDisplay()
	-- ABOUT THIS FUNCTION
	-- This is a very large part of DungeonMate - in fact, you can think of this function as the core of the addon.
	-- Essentially, this function will clear out all of the player frames shown in dm_panel and then regenerate it with
	-- (potentially) updated data. This function could possibly be called "often", though as a rule of thumb, try not to do
	-- anything that will cause it to get called more than once a minute on average. For instance, hooking this function into
	-- OnEvent would be very, very bad. Nevertheless, every effort should be made to make it run as quickly as possible.
	--
	-- In the future, it may become necessary to rework the entire control flow of the addon to make this function merely
	-- search out what players need updating and then update them, rather than rebuilding the entire list each call. For now,
	-- the 'wipe and rebuild' method works just as well.
	
	clearParty() -- Clears out the party UI, NOT the groupcache. We will rebuild the UI in this call.
	UpdateIgnorelist() -- Needs to be called at some point before data is harvested, so we call it here.
	ClearPartyGroupStatus() -- Loops over the groupcache marking everyone as Not In Group, so that later calls can update that.
	                        -- Note that I used to use UnitInGroup(unit), but it was being buggy for cross-realm units.
	
	-- FOR THE FOLLOWING TWO CALLS: The *only* updating either should do, is updating that can't be done elsewhere. Be
	-- VERY careful not to update any info in these calls that shouldn't be updated in a shared update routine, below.
	UpdatePartyMembers() -- This function makes sure that the current party members are all correctly accounted for
	UpdateNonPartyMembers() -- This function updates all people in the group cache that are NOT in the party. 
	
	UpdateGroupCache() -- Full-Group info gathering, things that all players can get wether in a group or not
	
	-- Information generation is now complete, so we can now start rebuilding the UI
	BuildGroupDisplay()
end

local function setPanelShown(isShown)
	if isShown then
		dm_panel:Show()
	else
		dm_panel:Hide()
	end
end

local function togglePanel()
	DungeonMate.db.char.show_panel = not DungeonMate.db.char.show_panel
	setPanelShown(DungeonMate.db.char.show_panel)
end

local function saveFramePositions()
		local point,rt,rp,xoff,yoff = plist:GetPoint(1)
		DungeonMate.db.char.panel_position.x = xoff
		DungeonMate.db.char.panel_position.y = yoff
		DungeonMate.db.char.panel_position.point = point
end


local function initializeFrame()
	dm_panel = CreateFrame("Frame",nil,UIParent)
	dm_panel:SetBackdrop({bgFile ="Interface/DialogFrame/UI-DialogBox-Background", 
						  edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
						  tile = true, tileSize = 16, edgeSize = 16, 
						  insets = { left = 4, right = 4, top = 4, bottom = 4 }})

	
	local title = dm_panel:CreateFontString(nil,"OVERLAY","GameFontHighlightLarge")
	title:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
	title:SetPoint("TOPLEFT",dm_panel,10,-7)
	title:SetText("DungeonMate")
	title:Show()
	
	--[[ FOOTER: Removed, but I might want it again in the future.
	local footer = CreateFrame("Frame",nil,dm_panel)
	footer:SetHeight(25)
	footer:Show()
	END Footer ]]--
	
	local close = CreateFrame("Button",nil,dm_panel,"UIPanelCloseButton")
	close:SetPoint("TOPRIGHT",dm_panel,"TOPRIGHT",3,3)
	close:SetScript("OnClick",togglePanel)
	close:SetHeight(30)
	close:SetWidth(30)

	plist = CreateFrame("Frame",nil,dm_panel)
	plist:SetPoint(DungeonMate.db.char.panel_position.point,UIParent,
	               DungeonMate.db.char.panel_position.point,
	               DungeonMate.db.char.panel_position.x,
	               DungeonMate.db.char.panel_position.y)
	plist:SetWidth(250)
	plist:SetHeight(4) -- dummy height, changes with added players
	plist.group = {}
	plist:Show()
	
	--[[ FOOTER: Removed, but I might want it again in the future.
	footer:SetPoint("BOTTOMLEFT",plist,"BOTTOMLEFT",0,-1*(footer:GetHeight()+4))
	footer:SetPoint("BOTTOMRIGHT",plist,"BOTTOMRIGHT",0,-1*(footer:GetHeight()+4))
	END Footer ]]--
	
	dm_panel:SetPoint("TOPLEFT",plist,"TOPLEFT",-10,20)
	dm_panel:SetPoint("BOTTOMRIGHT",plist,"BOTTOMRIGHT",10,-5)
	
	dm_panel:EnableMouse(true)
	plist:SetMovable(true)
	dm_panel:SetScript("OnMouseUp",function(frame)
		plist:StopMovingOrSizing()
		saveFramePositions()
	end)
	dm_panel:SetScript("OnMouseDown",function(frame,button)
		if button == "LeftButton" then
			plist:StartMoving()
		end
	end)
	dm_panel:SetScript("OnHide",function(frame)
		plist:StopMovingOrSizing()
		saveFramePositions()
	end)
	
	dm_panel:Hide()
end

----------[[ CATEGORY: INITIALIZATION ]]----------

function DungeonMate:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("DungeonMateDB")
	self.db:RegisterDefaults({
		char = {
			["show_panel"] = true,
			["panel_position"] = {
				["point"] = "CENTER",
				["x"] = 450,
				["y"] = 0,
			}
		},
		realm = {
			["pdb"] = {}
		}
	})
	self:RegisterChatCommand("dm",togglePanel)
	self:RegisterChatCommand("dungeonmate",togglePanel)
end

function DungeonMate:OnEnable()
	initializeFrame()
	initializeNoteEditFrame()
	setPanelShown(DungeonMate.db.char.show_panel)
	DungeonMate:RegisterEvent("PARTY_MEMBERS_CHANGED",UpdatePartyDisplay)
	lgt:RegisterCallback("LibGroupTalents_Update",UpdatePartyDisplay) 
	DungeonMate:RegisterEvent("IGNORELIST_UPDATE",UpdatePartyDisplay)
end

function DungeonMate:OnDisable()

end


----------[[ DATA BROKER ]]----------

local ldb = LibStub("LibDataBroker-1.1")

ldb:NewDataObject("DungeonMate", {
	type = "launcher",
	icon = dmicon,
	OnClick = togglePanel,
})

--- DEBUG FUNCTION ---
local dm_debug_n = 0
function dm_inject()
	dm_debug_n = dm_debug_n + 1
	local player = {
		["name"]="Random" .. dm_debug_n,
		["class"]="PRIEST",
		["classname"]="Priest",
		["level"]=80,
		["role"]="no role",
		["note"]="It's a trap!",
		["lastseentime"]=1262077608,
		["lastseenzone"]="Utgarde Keep",
	}
	groupcache[Random..dm_debug_n] = player
	UpdatePartyDisplay()
end